@@ -59,6 +59,7 @@
 //! - [RFC 2919 - List-Id: A Structured Field and Namespace for the Identification of Mailing Lists](https://datatracker.ietf.org/doc/html/rfc2919)
 //! - [RFC 3339 - Date and Time on the Internet: Timestamps](https://datatracker.ietf.org/doc/html/rfc3339)
 //! - [RFC 8621 - The JSON Meta Application Protocol (JMAP) for Mail (Section 4.1.4)](https://datatracker.ietf.org/doc/html/rfc8621#section-4.1.4)
+//! - [RFC 5957 - Internet Message Access Protocol - SORT and THREAD Extensions (Section 2.1)](https://datatracker.ietf.org/doc/html/rfc5256#section-2.1)
 //!
 //! ## Supported Character Sets
 //!
@@ -246,11 +247,7 @@
 //!
 //!    // Full RFC2231 support for continuations and character sets
 //!    assert_eq!(
-//!        nested_attachment
-//!            .get_content_type()
-//!            .unwrap()
-//!            .get_attribute("name")
-//!            .unwrap(),
+//!        nested_attachment.get_attachment_name().unwrap(),
 //!        "Book about ☕ tables.gif"
 //!    );
 //!
@@ -265,6 +262,7 @@ pub mod parsers;
 use std::{borrow::Cow, collections::HashMap, fmt};
 
 use decoders::html::{html_to_text, text_to_html};
+use parsers::fields::thread::thread_name;
 #[cfg(feature = "serde_support")]
 use serde::{Deserialize, Serialize};
 
@@ -324,6 +322,8 @@ pub struct Message<'x> {
     pub offset_header: usize,
     pub offset_body: usize,
     pub offset_end: usize,
+    #[cfg_attr(feature = "serde_support", serde(skip))]
+    pub raw_message: Cow<'x, [u8]>,
 }
 
 /// Body structure.
@@ -416,9 +416,24 @@ impl<'x> MessagePart<'x> {
         }
     }
 
-    pub fn unwrap_message(&self) -> Message {
+    pub fn unwrap_message(&self) -> &Message {
         match self {
-            MessagePart::Message(part) => part.parse().unwrap(),
+            MessagePart::Message(part) => match &part.body {
+                MessageAttachment::Parsed(message) => message.as_ref(),
+                MessageAttachment::Raw(_) => panic!(
+                    "This message part has not been parsed yet, use parse_message() instead."
+                ),
+            },
+            _ => panic!("Expected message part."),
+        }
+    }
+
+    pub fn parse_message(&'x self) -> Option<Message<'x>> {
+        match self {
+            MessagePart::Message(part) => match &part.body {
+                MessageAttachment::Parsed(_) => None,
+                MessageAttachment::Raw(raw_message) => Message::parse(raw_message.as_ref()),
+            },
             _ => panic!("Expected message part."),
         }
     }
@@ -870,6 +885,11 @@ impl<'x> Message<'x> {
             .and_then(|header| header.as_text_ref())
     }
 
+    /// Returns the message thread name
+    pub fn get_thread_name(&self) -> Option<&str> {
+        thread_name(self.get_subject()?).into()
+    }
+
     /// Returns the To header field
     pub fn get_to(&self) -> &HeaderValue<'x> {
         self.headers_rfc
@@ -983,6 +1003,15 @@ pub trait MimeHeaders<'x> {
     fn get_content_language(&self) -> &HeaderValue<'x>;
     /// Returns the Content-Location field
     fn get_content_location(&self) -> Option<&str>;
+    /// Returns the attachment name, if any.
+    fn get_attachment_name(&self) -> Option<&str> {
+        self.get_content_disposition()
+            .and_then(|cd| cd.get_attribute("filename"))
+            .or_else(|| {
+                self.get_content_type()
+                    .and_then(|ct| ct.get_attribute("name"))
+            })
+    }
 }
 
 impl<'x> MimeHeaders<'x> for Message<'x> {
@@ -1110,11 +1139,14 @@ impl<'x> fmt::Display for Part<'x, Cow<'x, [u8]>> {
 
 impl<'x> BodyPart<'x> for Part<'x, MessageAttachment<'x>> {
     fn get_contents(&'x self) -> &'x [u8] {
-        self.body.raw_message.as_ref()
+        match &self.body {
+            MessageAttachment::Parsed(msg) => msg.raw_message.as_ref(),
+            MessageAttachment::Raw(raw) => raw.as_ref(),
+        }
     }
 
     fn get_text_contents(&'x self) -> &'x str {
-        std::str::from_utf8(self.body.raw_message.as_ref()).unwrap_or("")
+        std::str::from_utf8(self.get_contents()).unwrap_or("")
     }
 
     fn is_text(&self) -> bool {
@@ -1126,13 +1158,16 @@ impl<'x> BodyPart<'x> for Part<'x, MessageAttachment<'x>> {
     }
 
     fn len(&self) -> usize {
-        self.body.raw_message.len()
+        match &self.body {
+            MessageAttachment::Parsed(msg) => msg.raw_message.len(),
+            MessageAttachment::Raw(raw) => raw.len(),
+        }
     }
 }
 
 impl<'x> fmt::Display for Part<'x, MessageAttachment<'x>> {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-        fmt.write_str(String::from_utf8_lossy(self.body.raw_message.as_ref()).as_ref())
+        fmt.write_str(self.get_text_contents())
     }
 }
 
@@ -1262,23 +1297,71 @@ impl<'x> ContentType<'x> {
     }
 }
 
-/// Raw contents of an e-mail message attachment.
-#[derive(Debug, PartialEq, Eq)]
-pub struct MessageAttachment<'x> {
-    pub raw_message: Cow<'x, [u8]>,
+/// Contents of an e-mail message attachment.
+#[derive(Debug, PartialEq)]
+pub enum MessageAttachment<'x> {
+    Parsed(Box<Message<'x>>),
+    Raw(Cow<'x, [u8]>),
+}
+
+impl<'x> Default for MessageAttachment<'x> {
+    fn default() -> Self {
+        MessageAttachment::Raw((&[] as &[u8]).into())
+    }
 }
 
 impl<'x> MessageAttachment<'x> {
     /// Parse the message attachment and return a `Message` object.
-    pub fn parse(&'x self) -> Option<Message<'x>> {
-        Message::parse(self.raw_message.as_ref())
+    pub fn parse_raw(&'x self) -> Option<Message<'x>> {
+        if let MessageAttachment::Raw(raw) = self {
+            Message::parse(raw.as_ref())
+        } else {
+            None
+        }
     }
+
+    /// Returns a reference to the parsed `Message` object.
+    pub fn as_ref(&self) -> Option<&Message> {
+        if let MessageAttachment::Parsed(message) = self {
+            message.as_ref().into()
+        } else {
+            None
+        }
+    }
+
+    /// Returns ```true``` is the message has been parsed.
+    /// Call `MessageAttachment::parse_raw` when this method returns ```false```,
+    /// otherwise use `MessageAttachment::as_ref`.
+    pub fn is_parsed(&self) -> bool {
+        matches!(self, MessageAttachment::Parsed(_))
+    }
+
+    /*/// Returns the parsed message if available, otherwise
+    /// parses the raw message and returns the results.
+    pub fn unwrap(&'x mut self) -> Option<Message<'x>> {
+        match self {
+            MessageAttachment::Parsed(message) => (*std::mem::take(&mut message).unwrap()).into(),
+            MessageAttachment::Raw(raw) => Message::parse(raw.as_ref()),
+        }
+    }*/
 }
 
 impl<'x> Part<'x, MessageAttachment<'x>> {
     /// Parse the message attachment and return a `Message` object.
-    pub fn parse(&'x self) -> Option<Message<'x>> {
-        self.body.parse()
+    pub fn parse_raw(&'x self) -> Option<Message<'x>> {
+        self.body.parse_raw()
+    }
+
+    /// Returns a reference to the parsed `Message` object.
+    pub fn as_ref(&self) -> Option<&Message> {
+        self.body.as_ref()
+    }
+
+    /// Returns ```true``` is the message has been parsed.
+    /// Call `MessageAttachment::parse_raw` when this method returns ```false```,
+    /// otherwise use `MessageAttachment::as_ref`.
+    pub fn is_parsed(&self) -> bool {
+        self.body.is_parsed()
     }
 }
 
@@ -1287,9 +1370,12 @@ impl<'x> Serialize for MessageAttachment<'x> {
     where
         S: serde::Serializer,
     {
-        self.parse()
-            .ok_or_else(|| serde::ser::Error::custom("Failed to parse message attachment."))?
-            .serialize(serializer)
+        match self {
+            MessageAttachment::Parsed(message) => message.serialize(serializer),
+            MessageAttachment::Raw(raw) => Message::parse(raw.as_ref())
+                .ok_or_else(|| serde::ser::Error::custom("Failed to parse message attachment."))?
+                .serialize(serializer),
+        }
     }
 }
 
