@@ -9,50 +9,59 @@
  * except according to those terms.
  */
 
-use std::collections::hash_map::Entry;
+use std::{borrow::Cow, collections::hash_map::Entry};
 
-use crate::{HeaderName, HeaderValue, Headers};
+use crate::{HeaderName, HeaderOffset, HeaderOffsetName, HeaderValue, OtherHeaders, RfcHeaders};
 
 use super::{
     fields::{
-        address::parse_address,
-        content_type::parse_content_type,
-        date::parse_date,
-        id::parse_id,
-        list::parse_comma_separared,
-        raw::{parse_and_ignore, parse_raw},
-        unstructured::parse_unstructured,
+        address::parse_address, content_type::parse_content_type, date::parse_date, id::parse_id,
+        list::parse_comma_separared, raw::parse_raw, unstructured::parse_unstructured,
     },
     message::MessageStream,
 };
 
 #[derive(Debug, PartialEq)]
 pub enum HeaderParserResult<'x> {
-    Supported(HeaderName),
-    Unsupported(&'x [u8]),
+    Rfc(HeaderName),
+    Other(Cow<'x, str>),
     Lf,
     Eof,
 }
 
-pub fn parse_headers<'x>(headers: &mut Headers<'x>, stream: &mut MessageStream<'x>) -> bool {
+pub fn parse_headers<'x>(
+    headers_rfc: &mut RfcHeaders<'x>,
+    mut headers_other: Option<&mut OtherHeaders<'x>>,
+    mut headers_offsets: Option<&mut Vec<HeaderOffset<'x>>>,
+    stream: &mut MessageStream<'x>,
+) -> bool {
     loop {
         let (bytes_read, result) = parse_header_name(&stream.data[stream.pos..]);
         stream.pos += bytes_read;
 
         match result {
-            HeaderParserResult::Supported(name) => {
+            HeaderParserResult::Rfc(name) => {
                 let (is_many, parser) = HDR_PARSER[name as usize];
 
+                let from_offset = stream.pos;
                 let value = parser(stream);
+                if let Some(headers_offsets) = headers_offsets.as_mut() {
+                    headers_offsets.push(HeaderOffset {
+                        name: HeaderOffsetName::Rfc(name),
+                        start: from_offset,
+                        end: stream.pos,
+                    });
+                }
+
                 if !value.is_empty() {
                     if is_many {
-                        match headers.entry(name) {
+                        match headers_rfc.entry(name) {
                             Entry::Occupied(mut e) => {
                                 if let HeaderValue::Collection(col) = e.get_mut() {
                                     col.push(value);
                                 } else {
                                     let old_value = e.remove();
-                                    headers.insert(
+                                    headers_rfc.insert(
                                         name,
                                         HeaderValue::Collection(vec![old_value, value]),
                                     );
@@ -63,11 +72,42 @@ pub fn parse_headers<'x>(headers: &mut Headers<'x>, stream: &mut MessageStream<'
                             }
                         }
                     } else {
-                        headers.insert(name, value);
+                        headers_rfc.insert(name, value);
+                    }
+                }
+            }
+            HeaderParserResult::Other(name) => {
+                if let Some(headers_other) = headers_other.as_mut() {
+                    let from_offset = stream.pos;
+                    let value = parse_raw(stream);
+                    if let Some(headers_offsets) = headers_offsets.as_mut() {
+                        headers_offsets.push(HeaderOffset {
+                            name: HeaderOffsetName::Other(name.clone()),
+                            start: from_offset,
+                            end: stream.pos,
+                        });
+                    }
+
+                    if !value.is_empty() {
+                        match headers_other.entry(name.clone()) {
+                            Entry::Occupied(mut e) => {
+                                if let HeaderValue::Collection(col) = e.get_mut() {
+                                    col.push(value);
+                                } else {
+                                    let old_value = e.remove();
+                                    headers_other.insert(
+                                        name,
+                                        HeaderValue::Collection(vec![old_value, value]),
+                                    );
+                                }
+                            }
+                            Entry::Vacant(e) => {
+                                e.insert(value);
+                            }
+                        }
                     }
                 }
             }
-            HeaderParserResult::Unsupported(_name) => parse_and_ignore(stream),
             HeaderParserResult::Lf => return true,
             HeaderParserResult::Eof => return false,
         }
@@ -98,14 +138,14 @@ pub fn parse_header_name(data: &[u8]) -> (usize, HeaderParserResult) {
                             let token_hash = token_hash - 4;
 
                             if field.eq_ignore_ascii_case(HDR_NAMES[token_hash]) {
-                                return (
-                                    bytes_read,
-                                    HeaderParserResult::Supported(HDR_MAP[token_hash]),
-                                );
+                                return (bytes_read, HeaderParserResult::Rfc(HDR_MAP[token_hash]));
                             }
                         }
                     }
-                    return (bytes_read, HeaderParserResult::Unsupported(field));
+                    return (
+                        bytes_read,
+                        HeaderParserResult::Other(String::from_utf8_lossy(field)),
+                    );
                 }
             }
             b'\n' => {
@@ -133,24 +173,11 @@ pub fn parse_header_name(data: &[u8]) -> (usize, HeaderParserResult) {
     (bytes_read, HeaderParserResult::Eof)
 }
 
-/*fn set_unsupported(&mut self, stream: &mut MessageStream<'x>, name: &'x [u8]) {
-    if let Ok(name) = std::str::from_utf8(name) {
-        let value = parse_unstructured(stream);
-        if !value.is_empty() {
-            match self.others.entry(name) {
-                Entry::Occupied(mut e) => {
-                    if let HeaderValue::Collection(col) = e.get_mut() {
-                        col.push(value);
-                    } else {
-                        let old_value = e.remove();
-                        self.others.insert(name,HeaderValue::Collection(vec![old_value, value]));
-                    }
-                },
-                Entry::Vacant(e) => {e.insert(value);},
-            }
-        }
+impl From<HeaderName> for u8 {
+    fn from(name: HeaderName) -> Self {
+        name as u8
     }
-}*/
+}
 
 #[cfg(test)]
 mod tests {
@@ -161,27 +188,24 @@ mod tests {
     #[test]
     fn header_name_parse() {
         let inputs = [
-            ("From: ", HeaderParserResult::Supported(HeaderName::From)),
-            (
-                "receiVED: ",
-                HeaderParserResult::Supported(HeaderName::Received),
-            ),
+            ("From: ", HeaderParserResult::Rfc(HeaderName::From)),
+            ("receiVED: ", HeaderParserResult::Rfc(HeaderName::Received)),
             (
                 " subject   : ",
-                HeaderParserResult::Supported(HeaderName::Subject),
+                HeaderParserResult::Rfc(HeaderName::Subject),
             ),
             (
                 "X-Custom-Field : ",
-                HeaderParserResult::Unsupported(b"X-Custom-Field"),
+                HeaderParserResult::Other("X-Custom-Field".into()),
             ),
-            (" T : ", HeaderParserResult::Unsupported(b"T")),
+            (" T : ", HeaderParserResult::Other("T".into())),
             (
                 "mal formed: ",
-                HeaderParserResult::Unsupported(b"mal formed"),
+                HeaderParserResult::Other("mal formed".into()),
             ),
             (
                 "MIME-version : ",
-                HeaderParserResult::Supported(HeaderName::MimeVersion),
+                HeaderParserResult::Rfc(HeaderName::MimeVersion),
             ),
         ];
 
@@ -235,7 +259,7 @@ static HDR_PARSER: &[(
     (false, parse_address),         // ListPost = 34,
     (false, parse_address),         // ListSubscribe = 35,
     (false, parse_address),         // ListUnsubscribe = 36,
-    (false, parse_raw),             // Unsupported = 37,
+    (true, parse_raw),              // Other = 37,
 ];
 
 static HDR_HASH: &[u8] = &[
@@ -254,18 +278,18 @@ static HDR_HASH: &[u8] = &[
 
 static HDR_MAP: &[HeaderName] = &[
     HeaderName::Date,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::Sender,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::Received,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::References,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::Cc,
     HeaderName::Comments,
     HeaderName::ResentCc,
     HeaderName::ContentId,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::ResentMessageId,
     HeaderName::ReplyTo,
     HeaderName::ResentTo,
@@ -273,54 +297,54 @@ static HDR_MAP: &[HeaderName] = &[
     HeaderName::ContentLanguage,
     HeaderName::Subject,
     HeaderName::ResentSender,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
+    HeaderName::Other,
+    HeaderName::Other,
     HeaderName::ResentDate,
     HeaderName::To,
     HeaderName::Bcc,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::ContentTransferEncoding,
     HeaderName::ReturnPath,
     HeaderName::ListId,
     HeaderName::Keywords,
     HeaderName::ContentDescription,
     HeaderName::ListOwner,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::ContentType,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::ListHelp,
     HeaderName::MessageId,
     HeaderName::ContentLocation,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
+    HeaderName::Other,
+    HeaderName::Other,
     HeaderName::ListSubscribe,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
     HeaderName::ListPost,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::ResentFrom,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
+    HeaderName::Other,
+    HeaderName::Other,
     HeaderName::ContentDisposition,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::InReplyTo,
     HeaderName::ListArchive,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::From,
-    HeaderName::Unsupported,
+    HeaderName::Other,
     HeaderName::ListUnsubscribe,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
-    HeaderName::Unsupported,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
+    HeaderName::Other,
     HeaderName::MimeVersion,
 ];
 
